// Paket mit bergeordneter Klasse fr die Darstellung am
// Bildschirm in Parallelprojektion, direkt auf GRX basierend
// Spezifikationen
// (c) 1996 by Andreas Meile, HTL Brugg-Windisch
// Klasse 3 Ia  Datum: 28.11.1996

#ifndef lib3d_h
#define lib3d_h

#include <grx.h>
#include "vektor3d.h"

enum obj_typ {kugel, drahtm, polyeder}; // Aufzhltyp fr Objektarten

typedef int int_pt[2];

// Hilfsfunktionen
void grafikmodus_ein(int hg=0, int vg=7);
void grafikmodus_aus();
int grafik_max_x();
int grafik_max_y();

class objekt3d {
  private:
    int uf, ff;      // Umriss- und Fllfarbe
    obj_typ typ;     // Objekttyp
    vektor3d m;      // Schwerpunkt fr Zeichenreihenfolge-Sort, damit
                     // mehrere Objekte in der korrekten Reihenfolge auf
                     // den Bildschirm gezeichnet werden
    int n_kno;       // Anzahl Eckpunkte
    int n_kno_max;   // Maximale Anzahl Knoten (fr Speicherfreigabe)
    vektor3d *pkt;   // Zeiger auf Feld mit Vektoren
    int_pt *bxy;     // Zeiger auf fr den Bildschirm umgerechnete Koord.
    int n_zdef;      // Anzahl Elemente fr Kurvenzugdefinitionen
    int n_zdef_max;  // Maximale Anzahl Kurvenzugdefinitionseintrge
    int *zdef;       // Zeiger auf Feld, das die Zeichendefinitionen enthlt
    objekt3d *naechst; // Zeiger auf nchstes 3D-Objekt

    friend class szene3d;
};

class szene3d {
  private:
    // Hilfsstruktur fr z-Sortierung
    struct z_sor_ptr {
      float z;       // z-Koordinate des Mittelpunktes
      objekt3d *obj; // Zeiger auf zugehriges Objekt
    };

    // Anker fr 3D-Objekt selber
    int hg;          // Hintergrundfarbe
    int n_obj;       // Anzahl Objekte
    int max_pts;     // Lngster Polygon-Linienzug
    objekt3d *erst;  // Zeiger auf das erste
    GrContext *puff; // Context-Puffer => der Grafikmodus muss bereits
    // aktiv sein, noch bevor das Objekt angelegt wird!

    // Ausgangsparameter fr Elementaroperationen
    vektor3d ev[3]; // Matrix aus drei Einheitsvektoren
    float xf, yf;   // Streckfaktoren fr den Bildschirm
    float xmb, ymb; // Mittelpunkt im Grafik-Context
    vektor3d mr;    // Mittelpunkt im Raum
    int xpos, ypos; // Position am Bildschirm
    int xmax, ymax; // Grsse des Context - 1

    z_sor_ptr *sortfeld; // Sortierfeld fr z-Koordinaten
    int_pt (*b_koord); // Feld fr Bildschirmkoordinaten

    void aktual_neuzeichnen();
  public:
    // Leere 3D-Objektszene einrichten
    szene3d ();
    // Wegen dynamischen Datenstrukturen auch ein Destruktor
    ~szene3d ();

    // Methoden fr den Benutzer: Elementaroperationen
    // Weiteres Objekt in Szenarion
    void neues_objekt(int n_knot,   // Anzahl Eckknoten
                      int n_kdefi,  // Anzahl Eckverbindungsdefinitionen
                const vektor3d &mi, // Mittelpunkt fr Zeichenreihenfolge-Sort
                      int umr_far,  // Umrissfarbe/Drahtfarbe
                      int f_far = -1, // Fllfarbe
                      obj_typ o_art = drahtm); // Objekttyp

    int fuege_knoten_ein(const vektor3d &p); // Knoten einfgen
    // Rckgabewert -1 = kein Platz mehr
    int fuege_k_zug_ein(int pnr[]);    // Feld mit Indizies: erster Eintrag =
    // Anzahl Punkte, der zweite Eintrag ist emntweder der Beginn der Liste
    // oder -1 fr aufsteigende Folge => dritter Eintrag ist dann Startindex
    // Beispiel: {3, 3, 2, 6} : Dreieck aus den Ecken 3, 2 und 6
    // Beispiel: {5, -1, 2} ist quivalent zu {5, 2, 3, 4, 5, 6} (Fnfeck)

    // Rckgabewert -1 = kein Speicherplatz mehr

    void aktualisiere_mittelpunkt();   // Mittelpunkt aktualisieren
  
    // Pfannenfertige Makros: Drahtmodell = Vorgabewerte
    void fuege_kugel_ein(const vektor3d &z,  // Kugelmittelpunkt
                                  float r,   // Radius
                              int umr_far,   // Umrissfarbe / Drahtfarbe
                              int f_far=0);  // Fllfarbe

    void fuege_pyramide_ein(const vektor3d &s, // Spitze
                            vektor3d &k, // Kantenvektor von Spitze zur
                            vektor3d &m, // Grundflche, Mittelachsenvektor
                            int n_ecken, // Anzahl Ecken
                            int umr_far, // Umrissfarbe
                          int f_far=-1); // Fllfarbe

    void fuege_prisma_ein(const vektor3d &s, // Erster Mittelpunkt
                          vektor3d &m, // Mittelachsenvektor (Betrag = Hhe)
                          vektor3d &r, // Radiusvektor
                          int n_ecken, // Anzahl Ecken
                          int umr_far, // Umrissfarbe
                        int f_far=-1); // Fllfarbe
  
    void fuege_parallelepiped_ein(vektor3d &s,  // Basispunkt
                                  vektor3d &k1, // 1. Kante
                            const vektor3d &k2, // 2. Kante
                            const vektor3d &k3, // 3. Kante
                                   int umr_far, // Umrissfarbe
                                 int f_far=-1); // Fllfarbe

    void fuege_tetraeder_ein(vektor3d &e1, // 1. Eckpunkt
                       const vektor3d &e2, // 2. Eckpunkt
                       const vektor3d &e3, // 3. Eckpunkt
                       const vektor3d &e4, // 4. Eckpunkt
                              int umr_far, // Umrissfarbe
                            int f_far=-1); // Fllfarbe
    
    // Objekt am Bildschirm anzeigen; dabei gleichzeitig
    // Sortier-Array, Grafik-Context etc. vorbereiten

    void zeige_an(const vektor3d &mitte,          // Ausschnittzentrum
      float bx, float by,                         // Breite und Hhe
      int h_grund = 0,                            // Hintergrundfarbe
      int bxmin = 0, int bxmax = grafik_max_x(),  // Ausschnitt des Bild-
      int bymin = 0, int bymax = grafik_max_y()); // schirms, um mit
      // mehreren solchen "szene3d"-Objekten die
      // Darstellung von mehreren Figuren gleichzeitig zu
      // ermglichen

    // Anzeige am Bildschirm wieder rckgngig machen
    // => Freigabe des Grafik-Contexts und des Sortier-Arrays
    void entferne_am_bildschirm();

    // Befehle zur Manipulation des Ausschnittes
    void reset();  // Zuweisen der Einheitsmatrix
    void korrektur(); // Feinorthonormierung der Transformationsmatrix:
    // Abweichungen infolge Rundungsfehler durch eine Orthonormalisierung
    // auskorrigieren
    void drehen(vektor3d &achse, float w); // Blickachse drehen
    void zoom_in(float f); // Vergrssern/Verkleinern
    int dyn_drehen(); // Fr tastaturgesteuertes ndern des Blickwinkels
};

#endif
